"""
    This file contains all unit tests for the monitor-rules-table in the database. (Corresponding to the file:
    'dashboard/database/monitor-rules.py')
    See __init__.py for how to run the test-cases.
"""

import unittest

from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, mean, \
    EXECUTION_TIMES, NAME

NAME2 = 'main2'
SUITE = 3


class TestDBTests(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()

    def test_get_tests(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import get_tests, Tests
        result = get_tests()
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].name, Tests(name=NAME).name)
        self.assertTrue(result[0].run)

    def test_add_test(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import add_test, get_tests
        self.test_get_tests()  # checks if there is 1 test in the DB
        add_test(NAME2)
        self.assertEqual(len(get_tests()), 2)

    def test_add_or_update_test(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import add_or_update_test, get_tests
        import datetime
        self.test_add_test()  # adds a new test
        self.assertEqual(NAME2, get_tests()[1].name)
        time = datetime.datetime.now()
        succeeded = True
        add_or_update_test(NAME2, time, succeeded)
        result = get_tests()
        self.assertEqual(len(result), 2)
        self.assertEqual(result[1].name, NAME2)
        self.assertEqual(result[1].lastRun, time)
        self.assertEqual(result[1].succeeded, succeeded)

    def test_add_test_result(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import get_results, add_test_result
        from flask_monitoringdashboard import config
        import datetime
        self.assertEqual(get_results(), [])
        for exec_time in EXECUTION_TIMES:
            add_test_result(NAME, exec_time, datetime.datetime.now(), config.version, SUITE, 2)
        result = get_results()
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].name, NAME)
        self.assertEqual(result[0].count, len(EXECUTION_TIMES))
        self.assertEqual(result[0].average, mean(EXECUTION_TIMES))

    def test_get_suite_nr(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import get_suite_nr, add_test_result
        from flask_monitoringdashboard import config
        import datetime
        self.assertEqual(get_suite_nr(), 1)
        add_test_result(NAME, 1234, datetime.datetime.now(), config.version, SUITE, 2)
        self.assertEqual(get_suite_nr(), SUITE+1)

    def test_get_results(self):
        """
            Test whether the function returns the right values.
        """
        self.test_add_test_result()  # can be replaced by test_add_test_result, since this function covers two tests

    def test_get_res_current(self):
        """
            Test whether the function returns the right values.
        """
        self.test_add_test_result()
        from flask_monitoringdashboard.database.tests import get_res_current
        from flask_monitoringdashboard import config
        result = get_res_current(config.version)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].name, NAME)
        self.assertEqual(result[0].count, len(EXECUTION_TIMES))
        self.assertEqual(result[0].average, mean(EXECUTION_TIMES))

        new_version = 'new_version'
        self.assertNotEqual(new_version, config.version)
        self.assertEqual(get_res_current(new_version), [])

    def test_get_line_results(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import get_line_results
        from flask_monitoringdashboard import config
        self.test_add_test_result()
        result = get_line_results()
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].version, config.version)
        self.assertEqual(result[0].avg, mean(EXECUTION_TIMES))
        self.assertEqual(result[0].min, min(EXECUTION_TIMES))
        self.assertEqual(result[0].max, max(EXECUTION_TIMES))
        self.assertEqual(result[0].count, len(EXECUTION_TIMES))

    def test_get_suites(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import get_suites
        self.test_add_test_result()
        self.assertEqual(get_suites(), [(SUITE, )])

    def test_get_measurements(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import get_measurements
        from flask_monitoringdashboard import config
        self.assertEqual(get_measurements(SUITE), [])
        self.test_add_test_result()
        result = get_measurements(SUITE)
        self.assertEqual(len(result), len(EXECUTION_TIMES))
        for test_run in result:
            self.assertEqual(test_run.name, NAME)
            self.assertIn(test_run.execution_time, EXECUTION_TIMES)
            self.assertEqual(test_run.version, config.version)
            self.assertEqual(test_run.suite, SUITE)

    def test_get_test_measurements(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import get_test_measurements
        from flask_monitoringdashboard import config
        self.assertEqual(get_test_measurements(NAME, SUITE), [])
        self.test_add_test_result()
        result = get_test_measurements(NAME, SUITE)
        self.assertEqual(len(result), len(EXECUTION_TIMES))
        for test_run in result:
            self.assertEqual(test_run.name, NAME)
            self.assertIn(test_run.execution_time, EXECUTION_TIMES)
            self.assertEqual(test_run.version, config.version)
            self.assertEqual(test_run.suite, SUITE)



from flask_monitoringdashboard import blueprint, config
from flask import redirect, request, session, render_template, url_for
from flask_monitoringdashboard.forms import Login
from flask_monitoringdashboard.security import check_login, on_logout


@blueprint.route('/login', methods=['GET', 'POST'])
def login():
    if session.get(config.link + '_logged_in'):
        return redirect(url_for('dashboard.index'))

    form = Login()
    if request.method == 'POST' and form.validate():
        if not check_login(name=form.name.data, password=form.password.data):
            form.name.errors.append('Incorrect username or password')
        else:
            return redirect(url_for('dashboard.index'))
    return render_template('dashboard/login.html', session=session, form=form)


@blueprint.route('/logout')
def logout():
    return on_logout()



"""
    This file can be executed for developing purposes. It is not used, when the flask_monitoring_dashboard is
    attached to an existing flask application.
"""

from flask import Flask, redirect, url_for


def create_app():
    import flask_monitoringdashboard as dashboard

    app = Flask(__name__)

    def get_session_id():
        # implement here your own custom function
        return '12345'

    dashboard.config.get_group_by = get_session_id
    dashboard.bind(app=app)

    @app.route('/')
    def main():
        return redirect(url_for('dashboard.index'))

    return app


if __name__ == '__main__':
    create_app().run(debug=True, host='0.0.0.0')



import datetime

import plotly
import plotly.graph_objs as go
from flask import session, render_template

from flask_monitoringdashboard import blueprint, config
from flask_monitoringdashboard.colors import get_color
from flask_monitoringdashboard.database.endpoint import get_last_accessed_times, get_num_requests
from flask_monitoringdashboard.database.function_calls import get_times, get_reqs_endpoint_day, get_versions, \
    get_data_per_version, get_endpoints, get_data_per_endpoint
from flask_monitoringdashboard.security import secure, is_admin


@blueprint.route('/measurements/overview')
@secure
def overview():
    colors = {}
    for result in get_times():
        colors[result.endpoint] = get_color(result.endpoint)
    return render_template('dashboard/measurement-overview.html', link=config.link, curr=2, times=get_times(),
                           colors=colors, access=get_last_accessed_times(), session=session, index=0,
                           is_admin=is_admin())


@blueprint.route('/measurements/heatmap')
@secure
def heatmap():
    colors = {}
    for result in get_times():
        colors[result.endpoint] = get_color(result.endpoint)
    return render_template('dashboard/measurement.html', link=config.link, curr=2, session=session, index=1,
                           graph=get_heatmap(end=None))


@blueprint.route('/measurements/requests')
@secure
def page_number_of_requests_per_endpoint():
    colors = {}
    for result in get_times():
        colors[result.endpoint] = get_color(result.endpoint)
    return render_template('dashboard/measurement.html', link=config.link, curr=2, session=session, index=2,
                           graph=get_stacked_bar())


@blueprint.route('/measurements/versions')
@secure
def page_boxplot_per_version():
    colors = {}
    for result in get_times():
        colors[result.endpoint] = get_color(result.endpoint)
    return render_template('dashboard/measurement.html', link=config.link, curr=2, session=session, index=3,
                           graph=get_boxplot_per_version())


@blueprint.route('/measurements/endpoints')
@secure
def page_boxplot_per_endpoint():
    colors = {}
    for result in get_times():
        colors[result.endpoint] = get_color(result.endpoint)
    return render_template('dashboard/measurement.html', link=config.link, curr=2, session=session, index=4,
                           graph=get_boxplot_per_endpoint())


def get_stacked_bar():
    data = get_reqs_endpoint_day()

    if len(data) == 0:
        return None

    labels = []
    endpoints = []
    # find labels and endpoints
    for d in data:
        if d.newTime not in labels:
            labels.append(d.newTime)
        if d.endpoint not in endpoints:
            endpoints.append(d.endpoint)
    labels.sort(reverse=True)

    # create dictionary. graph_data is in the form of: graph_data[label][endpoint]
    graph_data = {}
    for label in labels:
        graph_data[label] = {}
        for endpoint in endpoints:
            graph_data[label][endpoint] = 0

    # put data in dictionary
    for d in data:
        graph_data[d.newTime][d.endpoint] = d.cnt

    # create graph
    trace = []
    for endpoint in endpoints:

        lst = []
        for label in labels:
            lst.append(graph_data[label][endpoint])

        trace.append(go.Bar(
            y=labels,
            x=lst,
            name=endpoint,
            orientation='h',
            marker=dict(
                color=get_color(endpoint)
            )
        ))

    layout = go.Layout(
        barmode='stack',
        autosize=True,
        height=350 + 40 * len(labels),
        plot_bgcolor='rgba(249,249,249,1)',
        showlegend=True,
        title='Number of requests per endpoint per day',
        xaxis=dict(title='Number of requests'),
        yaxis=dict(autorange='reversed')
    )

    return plotly.offline.plot(go.Figure(data=trace, layout=layout), output_type='div', show_link=False)


def get_boxplot_per_version():
    """
    Creates a graph with the execution times per version
    :return:
    """
    versions = get_versions()

    if len(versions) == 0:
        return None

    data = []
    for v in versions:
        values = [c.execution_time for c in get_data_per_version(v.version)]
        data.append(go.Box(
            x=values,
            marker=dict(
                color=get_color(v.version)
            ),
            name="{0} {1}".format(v.version, v.startedUsingOn.strftime("%b %d %H:%M"))))

    layout = go.Layout(
        autosize=True,
        height=350 + 40 * len(versions),
        plot_bgcolor='rgba(249,249,249,1)',
        showlegend=False,
        title='Execution time for every version',
        xaxis=dict(title='Execution time (ms)'),
        yaxis=dict(autorange='reversed'),
        margin=go.Margin(
            l=200
        )
    )
    return plotly.offline.plot(go.Figure(data=data, layout=layout), output_type='div', show_link=False)


def get_boxplot_per_endpoint():
    """
    Creates a graph with the execution times per endpoint
    :return:
    """
    endpoints = [str(e.endpoint) for e in get_endpoints()]

    data = []
    for endpoint in endpoints:
        values = [c.execution_time for c in get_data_per_endpoint(endpoint)]
        if len(values) == 0:
            continue
        data.append(go.Box(
            x=values,
            name=endpoint,
            marker=dict(
                color=get_color(endpoint)
            )
        ))

    if len(data) == 0:
        return None

    layout = go.Layout(
        autosize=True,
        height=350 + 40 * len(endpoints),
        plot_bgcolor='rgba(249,249,249,1)',
        showlegend=False,
        title='Execution time for every endpoint',
        xaxis=dict(title='Execution time (ms)'),
        margin=go.Margin(
            l=200
        )
    )
    return plotly.offline.plot(go.Figure(data=data, layout=layout), output_type='div', show_link=False)


def get_heatmap(end):
    # list of hours: 1:00 - 23:00
    hours = ['0' + str(hour) + ':00' for hour in range(0, 10)] + \
            [str(hour) + ':00' for hour in range(10, 24)]

    data = get_num_requests(end)
    # list of days (format: year-month-day)
    days = [str(d.newTime[:10]) for d in data]
    # remove duplicates and sort the result
    days = sorted(list(set(days)))

    if len(data) == 0:
        return None

    first_day = max(datetime.datetime.strptime(days[0], '%Y-%m-%d'),
                    datetime.datetime.now() - datetime.timedelta(days=30))
    first_day -= datetime.timedelta(hours=12)
    last_day = datetime.datetime.strptime(days[len(days) - 1], '%Y-%m-%d') + datetime.timedelta(hours=12)

    # create empty 2D-dictionary with the keys: [hour][day]
    requests = {}
    for hour in hours:
        requests_day = {}
        for day in days:
            requests_day[day] = 0
        requests[hour] = requests_day

    # add data to the dictionary
    for d in data:
        day = str(d.newTime[:10])
        hour = str(d.newTime[11:16])
        requests[hour][day] = d.count

    # create a 2D-list out of the dictionary
    requests_list = []
    for hour in hours:
        day_list = []
        for day in days:
            day_list.append(requests[hour][day])
        requests_list.append(day_list)

    layout = go.Layout(
        autosize=True,
        height=800,
        plot_bgcolor='rgba(249,249,249,1)',
        showlegend=False,
        title='Heatmap of number of requests',
        xaxis=go.XAxis(range=[first_day, last_day],
                       title='Date'),
        yaxis=dict(title='Time', autorange='reversed')
    )

    trace = go.Heatmap(z=requests_list, x=days, y=hours)
    return plotly.offline.plot(go.Figure(data=[trace], layout=layout), output_type='div', show_link=False)



"""
    Create a StackInfo-object and it will spawn a different thread.
    That thread will sleep for 'average' ms (reformatted to seconds).
    Then it will log the stacktrace of all active threads.
    Moreover, it logs cpu- and memory-info.
"""

import time
import traceback
from threading import Thread
import sys
import psutil


class StackInfo(object):
    def __init__(self, average):
        self.average = average
        self.stacktrace = ''
        self.cpu_percent = ''
        self.memory = ''

        try:
            thread = Thread(target=log_stack_trace, args=(self,))
            thread.start()
        except Exception as e:
            print('Can\'t log traceback information: {}'.format(e))
            traceback.print_exc()


def log_stack_trace(stack_info):
    # average is in ms, sleep requires seconds
    time.sleep(stack_info.average / 1000.0)

    # iterate through every active thread and store the stack-trace
    stack_list = []
    for thread_id, stack in sys._current_frames().items():
        stack_list.append("\n# Thread_id: %s" % thread_id)
        for filename, lineno, name, line in traceback.extract_stack(stack):
            stack_list.append('File: "%s", line %d, in %s' % (filename, lineno, name))
            if line:
                stack_list.append("  %s" % (line.strip()))

    # Set the values in the object
    stack_info.stacktrace = '<br />'.join(stack_list)
    stack_info.cpu_percent = str(psutil.cpu_percent(interval=None, percpu=True))
    stack_info.memory = str(psutil.virtual_memory())



import datetime
from flask import json
from flask_monitoringdashboard.database import session_scope, Outlier


def add_outlier(endpoint, execution_time, stack_info, request):
    """ Collects information (request-parameters, memory, stacktrace) about the request and adds it in the database. """
    with session_scope() as db_session:
        outlier = Outlier(endpoint=endpoint, request_values=json.dumps(request.values),
                          request_headers=str(request.headers), request_environment=str(request.environ),
                          request_url=str(request.url), cpu_percent=stack_info.cpu_percent,
                          memory=stack_info.memory, stacktrace=stack_info.stacktrace,
                          execution_time=execution_time, time=datetime.datetime.now())
        db_session.add(outlier)


def get_outliers(endpoint):
    """ Returns a list of all outliers of a specific endpoint. """
    with session_scope() as db_session:
        result = db_session.query(Outlier).filter(Outlier.endpoint == endpoint).all()
        db_session.expunge_all()
        return result



from flask_wtf import FlaskForm
from wtforms import validators, SubmitField, PasswordField, StringField


class MonitorDashboard(FlaskForm):
    """ Used for selecting a number of checkboxes that corresponds to which endpoints should be followed. 
    Since the number of checkboxes is variable (equals the number of endpoints), none of them is added as
    a field to this class. In fact, they are generated in the template in plain html. """


class Login(FlaskForm):
    """ Used for serving a login form. """
    name = StringField('Username', [validators.data_required()])
    password = PasswordField('Password', [validators.data_required()])
    submit = SubmitField('Login')


class ChangeSetting(FlaskForm):
    """ Used for changing the username or password that is required to login on the dashboard. """
    username = StringField('Username', [validators.data_required()])
    password = PasswordField('New password', [validators.equal_to(fieldname='confirm')])
    confirm = PasswordField('Confirm password')
    submit = SubmitField('Update changes')


class RunTests(FlaskForm):
    """ Used for serving a login form on /{{ link }}/testmonitor. """
    submit = SubmitField('Run selected tests')



import unittest

from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, login, get_test_app


class TestLogin(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()
        self.app = get_test_app()

    def test_login(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        self.assertEqual(200, self.app.get('dashboard/login').status_code)
        login(self.app)
        self.assertEqual(302, self.app.get('dashboard/login').status_code)



"""
Contains all functions that operate on the testsGrouped table
"""
from flask_monitoringdashboard.database import session_scope, TestsGrouped


def reset_tests_grouped():
    """ Resets the testsGrouped table of the database. """
    with session_scope() as db_session:
        db_session.query(TestsGrouped).delete()


def add_tests_grouped(json):
    """ Adds endpoint - unit tests combinations to the database. """
    with session_scope() as db_session:
        for combination in json:
            db_session.add(TestsGrouped(endpoint=combination['endpoint'], test_name=combination['test_name']))


def get_tests_grouped():
    """ Return all existing endpoint - unit tests combinations. """
    with session_scope() as db_session:
        result = db_session.query(TestsGrouped).all()
        db_session.expunge_all()
        return result



"""
    Contains all functions that are used to track the performance of the flask-application.
    See init_measurement() for more detailed info.
"""
import time
import datetime
from flask import request
from functools import wraps
from flask_monitoringdashboard import user_app, config
from flask_monitoringdashboard.database.monitor_rules import get_monitor_rules
from flask_monitoringdashboard.database.endpoint import update_last_accessed
from flask_monitoringdashboard.database.function_calls import add_function_call
from flask_monitoringdashboard.database.outlier import add_outlier
from flask_monitoringdashboard.outlier import StackInfo

# count and sum are dicts and used for calculating the averages
endpoint_count = {}
endpoint_sum = {}


def init_measurement():
    """
    This should be added to the list of functions that are executed before the first request.
    This function is used in the config-method in __init__ of this folder
    It adds wrappers to the endpoints for tracking their performance and last access times.
    """
    for rule in get_monitor_rules():
        # add a wrapper for every endpoint
        if rule.endpoint in user_app.view_functions:
            user_app.view_functions[rule.endpoint] = track_performance(user_app.view_functions[rule.endpoint],
                                                                       endpoint=rule.endpoint)

    # filter dashboard rules
    rules = user_app.url_map.iter_rules()
    rules = [r for r in rules if not r.rule.startswith('/' + config.link)
             and not r.rule.startswith('/static-' + config.link)]
    for rule in rules:
        user_app.view_functions[rule.endpoint] = track_last_accessed(user_app.view_functions[rule.endpoint],
                                                                     endpoint=rule.endpoint)


def track_performance(func, endpoint):
    """
    Measure the execution time of a function and store result in the database
    :param func: the function to be measured
    :param endpoint: the name of the endpoint
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        # compute average
        average = get_average(endpoint)
        if average:
            average *= config.outlier_detection_constant

            # start a thread to log the stacktrace after 'average' ms
            stack_info = StackInfo(average)

        time1 = time.time()
        result = func(*args, **kwargs)
        time2 = time.time()
        t = (time2 - time1) * 1000
        add_function_call(time=t, endpoint=endpoint, ip=request.environ['REMOTE_ADDR'])

        # outlier detection
        endpoint_count[endpoint] += 1
        endpoint_sum[endpoint] += t
        # check for being an outlier
        if average and float(t) > average:
            add_outlier(endpoint, t, stack_info, request)

        return result

    wrapper.original = func
    return wrapper


def track_last_accessed(func, endpoint):
    """
    Keep track of the last access time of the endpoints. 
    :param func: the function to be measured
    :param endpoint: the name of the endpoint
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        update_last_accessed(endpoint=endpoint, value=datetime.datetime.now())
        return func(*args, **kwargs)

    return wrapper


def get_average(endpoint):
    if endpoint in endpoint_count:
        if endpoint_count[endpoint] < 10:
            return None
    else:
        # initialize endpoint
        endpoint_count[endpoint] = 0
        endpoint_sum[endpoint] = 0
        return None
    return endpoint_sum[endpoint] / endpoint_count[endpoint]



"""
    This file contains all unit tests for the endpoint-table in the database. (Corresponding to the file:
    'dashboard/database/endpoint.py')
    See __init__.py for how to run the test-cases.
"""

import unittest

from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, mean, \
    EXECUTION_TIMES, TIMES, NAME, GROUP_BY, IP


class TestEndpoint(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()

    def test_get_line_results(self):
        """
            Test whether the function returns the right values.
            It should return the average, min and max and length (count).
        """
        from flask_monitoringdashboard.database.endpoint import get_line_results
        result = get_line_results(NAME)[0]
        self.assertEqual(result.avg, mean(EXECUTION_TIMES))
        self.assertEqual(result.min, min(EXECUTION_TIMES))
        self.assertEqual(result.max, max(EXECUTION_TIMES))
        self.assertEqual(result.count, len(EXECUTION_TIMES))

    def test_get_num_requests(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.endpoint import get_num_requests
        result = get_num_requests(NAME)[0]
        self.assertEqual(result.count, len(EXECUTION_TIMES))

    def test_get_endpoint_column(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.endpoint import get_endpoint_column
        from flask_monitoringdashboard.database import FunctionCall
        result = get_endpoint_column(NAME, FunctionCall.time)
        self.assertEqual(len(result), len(EXECUTION_TIMES))
        self.assertEqual(result[0].time, TIMES[0])

    def test_get_endpoint_column_user_sorted(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.endpoint import get_endpoint_column_user_sorted
        from flask_monitoringdashboard.database import FunctionCall
        result = get_endpoint_column_user_sorted(NAME, FunctionCall.time)
        self.assertEqual(len(result), len(EXECUTION_TIMES))
        self.assertEqual(result[0].time, TIMES[0])

    def test_get_endpoint_results(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.endpoint import get_endpoint_results
        from flask_monitoringdashboard.database import FunctionCall
        result = get_endpoint_results(NAME, FunctionCall.group_by)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].average, mean(EXECUTION_TIMES))
        self.assertEqual(result[0].count, len(EXECUTION_TIMES))

    def test_get_monitor_rule(self):
        """
            Test wheter the function returns the right values.
        """
        from flask_monitoringdashboard.database.endpoint import get_monitor_rule
        from flask_monitoringdashboard import config
        rule = get_monitor_rule(NAME)
        self.assertEqual(rule.endpoint, NAME)
        self.assertTrue(rule.monitor)
        self.assertEqual(rule.version_added, config.version)

    def test_update_monitor_rule(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.endpoint import get_monitor_rule, update_monitor_rule
        current_value = get_monitor_rule(NAME).monitor
        new_value = not current_value
        update_monitor_rule(NAME, new_value)
        self.assertEqual(get_monitor_rule(NAME).monitor, new_value)

    def test_get_all_measurement(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard import config
        from flask_monitoringdashboard.database.endpoint import get_all_measurement
        result = get_all_measurement(NAME)
        self.assertEqual(len(result), len(EXECUTION_TIMES))
        for row in result:
            self.assertIn(row.execution_time, EXECUTION_TIMES)
            self.assertEqual(row.version, config.version)
            self.assertEqual(row.endpoint, NAME)
            self.assertEqual(row.group_by, GROUP_BY)
            self.assertEqual(row.ip, IP)

    def test_get_all_measurement_per_column(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.endpoint import get_all_measurement_per_column
        from flask_monitoringdashboard.database import FunctionCall
        from flask_monitoringdashboard import config
        result = get_all_measurement_per_column(endpoint=NAME, column=FunctionCall.ip, value=IP)
        self.assertEqual(len(result), len(EXECUTION_TIMES))
        for row in result:
            self.assertIn(row.execution_time, EXECUTION_TIMES)
            self.assertEqual(row.version, config.version)
            self.assertEqual(row.endpoint, NAME)
            self.assertEqual(row.group_by, GROUP_BY)
            self.assertEqual(row.ip, IP)

    def test_get_last_accessed_times(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.endpoint import get_last_accessed_times
        result = get_last_accessed_times()
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].endpoint, NAME)

    def test_update_last_accessed(self):
        """
            Test whether the function returns the right values.
        """
        import datetime
        time = datetime.datetime.now()
        from flask_monitoringdashboard.database.endpoint import update_last_accessed, get_last_accessed_times
        update_last_accessed(NAME, time)
        result = get_last_accessed_times()
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].endpoint, NAME)
        self.assertEqual(result[0].last_accessed, time)



import requests
import configparser
import time
import datetime
import os
import sys
import csv
from unittest import TestLoader

# Abort if config file is not specified.
config = os.getenv('DASHBOARD_CONFIG')
if config is None:
    print('You must specify a config file for the dashboard to be able to use the unit test monitoring functionality.')
    print('Please set an environment variable \'DASHBOARD_CONFIG\' specifying the absolute path to your config file.')
    sys.exit(0)

# Abort if log directory is not specified.
log_dir = os.getenv('DASHBOARD_LOG_DIR')
if log_dir is None:
    print('You must specify a log directory for the dashboard to be able to use the unit test monitoring '
          'functionality.')
    print('Please set an environment variable \'DASHBOARD_LOG_DIR\' specifying the absolute path where you want the '
          'log files to be placed.')
    sys.exit(0)

n = 1
url = None
sys.path.insert(0, os.getcwd())
parser = configparser.RawConfigParser()
try:
    parser.read(config)
    if parser.has_option('dashboard', 'N'):
        n = int(parser.get('dashboard', 'N'))
    if parser.has_option('dashboard', 'TEST_DIR'):
        test_dir = parser.get('dashboard', 'TEST_DIR')
    else:
        print('No test directory specified in your config file. Please do so.')
        sys.exit(0)
    if parser.has_option('dashboard', 'SUBMIT_RESULTS_URL'):
        url = parser.get('dashboard', 'SUBMIT_RESULTS_URL')
    else:
        print('No url specified in your config file for submitting test results. Please do so.')
except configparser.Error as e:
    print("Something went wrong while parsing the configuration file:\n{}".format(e))

data = {'test_runs': [], 'grouped_tests': []}
log = open(log_dir + "endpoint_hits.log", "w")
log.write("\"time\",\"endpoint\"\n")
log.close()
log = open(log_dir + "test_runs.log", "w")
log.write("\"start_time\",\"stop_time\",\"test_name\"\n")

if test_dir:
    suites = TestLoader().discover(test_dir, pattern="*test*.py")
    for i in range(n):
        for suite in suites:
            for case in suite:
                for test in case:
                    result = None
                    t1 = str(datetime.datetime.now())
                    time1 = time.time()
                    result = test.run(result)
                    time2 = time.time()
                    t2 = str(datetime.datetime.now())
                    log.write("\"{}\",\"{}\",\"{}\"\n".format(t1, t2, str(test)))
                    t = (time2 - time1) * 1000
                    data['test_runs'].append({'name': str(test), 'exec_time': t, 'time': str(datetime.datetime.now()),
                                              'successful': result.wasSuccessful(), 'iter': i + 1})

log.close()

# Read and parse the log containing the test runs
runs = []
with open(log_dir + 'test_runs.log') as log:
    reader = csv.DictReader(log)
    for row in reader:
        runs.append([datetime.datetime.strptime(row["start_time"], "%Y-%m-%d %H:%M:%S.%f"),
                     datetime.datetime.strptime(row["stop_time"], "%Y-%m-%d %H:%M:%S.%f"),
                     row['test_name']])

# Read and parse the log containing the endpoint hits
hits = []
with open(log_dir + 'endpoint_hits.log') as log:
    reader = csv.DictReader(log)
    for row in reader:
        hits.append([datetime.datetime.strptime(row["time"], "%Y-%m-%d %H:%M:%S.%f"),
                     row['endpoint']])

# Analyze logs to find out which endpoints are hit by which unit tests
for h in hits:
    for r in runs:
        if r[0] <= h[0] <= r[1]:
            if {'endpoint': h[1], 'test_name': r[2]} not in data['grouped_tests']:
                data['grouped_tests'].append({'endpoint': h[1], 'test_name': r[2]})
            break

# Try to send test results and endpoint-grouped unit tests to the flask_monitoringdashboard
if url:
    try:
        requests.post(url, json=data)
        print('Sent unit test results to the dashboard.')
    except Exception as e:
        print('Sending unit test results to the dashboard failed:\n{}'.format(e))



from flask import session, request, render_template

from flask_monitoringdashboard import blueprint, config, user_app
from flask_monitoringdashboard.database.endpoint import get_monitor_rule, update_monitor_rule, get_last_accessed_times
from flask_monitoringdashboard.database.monitor_rules import reset_monitor_endpoints, get_monitor_names
from flask_monitoringdashboard.database.tests import get_tests, get_results, get_suites, get_test_measurements
from flask_monitoringdashboard.database.tests import get_res_current, get_measurements
from flask_monitoringdashboard.database.tests_grouped import get_tests_grouped
from flask_monitoringdashboard.forms import MonitorDashboard
from flask_monitoringdashboard.measurement import track_performance
from flask_monitoringdashboard.security import secure, admin_secure
from flask_monitoringdashboard.colors import get_color

import plotly
import plotly.graph_objs as go


@blueprint.route('/settings', methods=['GET', 'POST'])
@admin_secure
def settings():
    password = 'x' * len(config.password)
    return render_template('dashboard/settings.html', link=config.link, session=session, config=config, pw=password)


def formatter(x):
    sec = x // 1000
    ms = round(x % 1000, 2)
    if sec == 0:
        return '{0}ms'.format(ms)
    return '{0}s and {1}ms'.format(sec, ms)


@blueprint.route('/rules', methods=['GET', 'POST'])
@admin_secure
def rules():
    form = MonitorDashboard()
    values = {}
    all_rules = user_app.url_map.iter_rules()

    if request.method == 'POST' and form.validate():
        # Remove the monitor endpoints from the database
        reset_monitor_endpoints()

        for rule in user_app.url_map.iter_rules():
            # Remove existing wrappers
            original = getattr(user_app.view_functions[rule.endpoint], 'original', None)
            if original:
                user_app.view_functions[rule.endpoint] = original

        # request.form only contains checkboxes that are checked
        for data in request.form:
            if data.startswith('checkbox-'):
                endpoint = data.rsplit('-', 1)[1]
                update_monitor_rule(endpoint, value=True)
                rule = get_monitor_rule(endpoint)
                # Add wrappers to the existing functions
                user_app.view_functions[rule.endpoint] = track_performance(user_app.view_functions[rule.endpoint],
                                                                           rule.endpoint)

    # store the result from the database in values (used for rendering)
    for rule in user_app.url_map.iter_rules():
        values[rule.endpoint] = get_monitor_rule(rule.endpoint).monitor

    la = get_last_accessed_times()

    # filter flask_monitoringdashboard rules
    all_rules = [r for r in all_rules if not r.rule.startswith('/' + config.link)
                 and not r.rule.startswith('/static-' + config.link)]
    colors = {}
    for rule in all_rules:
        colors[rule.endpoint] = get_color(rule.endpoint)

    return render_template('dashboard/rules.html', link=config.link, curr=1, rules=all_rules, access=la, form=form,
                           session=session, values=values, colors=colors)


@blueprint.route('/testmonitor/<test>')
@secure
def test_result(test):
    return render_template('dashboard/testresult.html', link=config.link, session=session, name=test,
                           boxplot=get_boxplot(test))


@blueprint.route('/testmonitor')
@secure
def testmonitor():
    endp_names = []
    for name in get_monitor_names():
        endp_names.append(name.endpoint)
    tests = get_tests_grouped()
    grouped = {}
    cols = {}
    for t in tests:
        if t.endpoint not in grouped:
            grouped[t.endpoint] = []
            cols[t.endpoint] = get_color(t.endpoint)
        if t.test_name not in grouped[t.endpoint]:
            grouped[t.endpoint].append(t.test_name)

    return render_template('dashboard/testmonitor.html', link=config.link, session=session, curr=3,
                           tests=get_tests(), endpoints=endp_names, results=get_results(), groups=grouped,
                           colors=cols, res_current_version=get_res_current(config.version), boxplot=get_boxplot(None))


def get_boxplot(test):
    data = []
    suites = get_suites()
    if not suites:
        return None
    for s in suites:
        if test:
            values = [str(c.execution_time) for c in get_test_measurements(name=test, suite=s.suite)]
        else:
            values = [str(c.execution_time) for c in get_measurements(suite=s.suite)]

        data.append(go.Box(
            x=values,
            name="{0} ({1})".format(s.suite, len(values))))

    layout = go.Layout(
        autosize=True,
        height=350 + 40 * len(suites),
        plot_bgcolor='rgba(249,249,249,1)',
        showlegend=False,
        title='Execution times for every Travis build',
        xaxis=dict(title='Execution time (ms)'),
        yaxis=dict(
            title='Build (measurements)',
            autorange='reversed'
        )
    )

    return plotly.offline.plot(go.Figure(data=data, layout=layout), output_type='div', show_link=False)
